"""more characters for settings and term column in station_ma_raster

Revision ID: V1.0.2
Revises: V1.0.0
Create Date: 2024-11-06 14:33:30.129005

"""
from typing import Sequence, Union

from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision: str = 'V1.0.2'
down_revision: Union[str, None] = 'V1.0.0'
branch_labels: Union[str, Sequence[str], None] = None
depends_on: Union[str, Sequence[str], None] = None


def upgrade() -> None:
    op.alter_column('settings', 'value',
               existing_type=sa.VARCHAR(length=20),
               type_=sa.String(length=60),
               existing_comment='The value of the setting',
               existing_nullable=False)

    op.execute(sa.text("DROP VIEW IF EXISTS station_ma_quotient_view CASCADE;"))
    op.drop_constraint("station_ma_raster_pkey", "station_ma_raster")
    op.add_column(
        'station_ma_raster',
        sa.Column('term',
                  sa.VARCHAR(length=4),
                  nullable=False,
                  server_default='year',
                  comment="The term of the raster. e.g. 'year', 'wihy', 'suhy'"))
    op.alter_column('station_ma_raster', 'value',
               existing_type=sa.INTEGER(),
               comment='The value of the raster for the station in the database unit.',
               existing_comment='The value of the raster for the station.',
               existing_nullable=False)
    op.execute(sa.text(
        """
        UPDATE public."station_ma_raster"
            SET
                term= split_part("parameter", '_', 2),
                parameter= split_part("parameter", '_', 1)
        WHERE parameter LIKE '%\_%';
        """))
    op.alter_column(
        'station_ma_raster', 'term',
        existing_server_default="fill", server_default=None)
    op.alter_column('station_ma_raster', 'parameter',
        existing_type=sa.VARCHAR(length=7),
        type_=sa.VARCHAR(length=3),
        comment="The parameter of the raster. e.g. 'p', 't', 'et'",
        existing_comment="The parameter of the raster. e.g. 'p_wihj', 'p_sohj', 'p_year', 't_year', 'et_year'",
        existing_nullable=False)
    op.create_primary_key("station_ma_raster_pkey", "station_ma_raster", ["station_id", "parameter", "raster_key", "term"])

def downgrade() -> None:
    op.alter_column('settings', 'value',
               existing_type=sa.String(length=60),
               type_=sa.VARCHAR(length=20),
               existing_comment='The value of the setting',
               existing_nullable=False)

    op.execute(sa.text("DROP VIEW IF EXISTS station_ma_timeseries_quotient_view CASCADE;"))
    op.drop_constraint("station_ma_raster_pkey", "station_ma_raster")
    op.alter_column('station_ma_raster', 'parameter',
               existing_type=sa.VARCHAR(length=3),
               type_=sa.VARCHAR(length=7),
               comment="The parameter of the raster. e.g. 'p_wihj', 'p_sohj', 'p_year', 't_year', 'et_year'",
               existing_comment="The parameter of the raster. e.g. 'p', 't', 'et'",
               existing_nullable=False)
    op.alter_column('station_ma_raster', 'value',
               existing_type=sa.INTEGER(),
               comment='The value of the raster for the station.',
               existing_comment='The value of the raster for the station in the database unit.',
               existing_nullable=False)
    op.execute(sa.text(
        """
        UPDATE public."station_ma_raster"
            SET "parameter" = "parameter"||'_'||"term"
        WHERE parameter NOT LIKE '%\_%';
        """))
    op.drop_column('station_ma_raster', 'term')
    op.create_primary_key("station_ma_raster_pkey", "station_ma_raster", ["station_id", "raster_key", "parameter"])



# libraries
import logging

from ..db.connections import db_engine
from ..station import StationP
from .StationsBase import StationsBase

# set settings
# ############
__all__ = ["StationsP"]
log = logging.getLogger(__name__)

# class definition
##################
class StationsP(StationsBase):
    """A class to work with and download 10 minutes precipitation data for several stations."""
    _StationClass = StationP
    _timeout_raw_imp = 360

    @db_engine.deco_update_privilege
    def update_richter_class(self, stids="all", do_mp=True, **kwargs):
        """Update the Richter exposition class.

        Get the value from the raster, compare with the richter categories and save to the database.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs : dict, optional
            The keyword arguments to be handed to the station.StationP.update_richter_class and get_stations method.

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        self._run_method(
            stations=self.get_stations(only_real=True, stids=stids, **kwargs),
            method="update_richter_class",
            name="update richter class for {para}".format(para=self._para.upper()),
            kwds=kwargs,
            do_mp=do_mp)

    @db_engine.deco_update_privilege
    def richter_correct(self, stids="all", **kwargs):
        """Richter correct the filled data.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        self._run_method(
            stations=self.get_stations(only_real=False, stids=stids, **kwargs),
            method="richter_correct",
            name="richter correction on {para}".format(para=self._para.upper()),
            do_mp=False, **kwargs)

    @db_engine.deco_update_privilege
    def last_imp_corr(self, stids="all", do_mp=False, **kwargs):
        """Richter correct the filled data for the last imported period.

        Parameters
        ----------
        stids: string or list of int, optional
            The Stations for which to compute.
            Can either be "all", for all possible stations
            or a list with the Station IDs.
            The default is "all".
        do_mp : bool, optional
            Should the method be done in multiprocessing mode?
            If False the methods will be called in threading mode.
            Multiprocessing needs more memory and a bit more initiating time. Therefor it is only usefull for methods with a lot of computation effort in the python code.
            If the most computation of a method is done in the postgresql database, then threading is enough to speed the process up.
            The default is False.
        **kwargs : dict, optional
            The additional keyword arguments for the _run_method and get_stations method

        Raises
        ------
        ValueError
            If the given stids (Station_IDs) are not all valid.
        """
        stations = self.get_stations(only_real=True, stids=stids, **kwargs)
        period = stations[0].get_last_imp_period(all=True)
        log.info("The {para_long} Stations fillup of the last import is started for the period {min_tstp} - {max_tstp}".format(
            para_long=self._para_long,
            **period.get_sql_format_dict(format="%Y%m%d %H:%M")))
        self._run_method(
            stations=stations,
            method="last_imp_corr",
            kwds={"_last_imp_period": period},
            name="richter correction on {para}".format(para=self._para.upper()),
            do_mp=do_mp, **kwargs)

    @db_engine.deco_update_privilege
    def update(self, only_new=True, **kwargs):
        """Make a complete update of the stations.

        Does the update_raw, quality check, fillup and richter correction of the stations.

        Parameters
        ----------
        only_new : bool, optional
            Should a only new values be computed?
            If False: The stations are updated for the whole possible period.
            If True, the stations are only updated for new values.
            The default is True.
        """
        super().update(only_new=only_new, **kwargs)
        if only_new:
            self.last_imp_richter_correct(**kwargs)
        else:
            self.richter_correct(**kwargs)



"""
This submodule has only one class Broker. This one is used to do actions on all the stations together. Mainly only used for updating the DB.
"""
# libraries
import logging
from sqlalchemy import text as sqltxt
from packaging import version as pv
from pathlib import Path
import textwrap
from contextlib import contextmanager
import atexit
import sys

from .db.connections import db_engine
from .stations import StationsP, StationsPD, StationsT, StationsET
from . import __version__

__all__ = ["Broker"]

log = logging.getLogger(__name__)

class Broker(object):
    """A class to manage and update the database.

    Can get used to update all the stations and parameters at once.

    This class is only working with SELECT and UPDATE user privileges. Even Better is to also have DELETE and INSERT privileges.
    """
    def __init__(self):
        if not db_engine.update_privilege & db_engine.select_privilege:
            raise PermissionError("You don't have enough privileges to use the Broker. The database user must have SELECT and UPDATE privileges.")

        self.stations_pd = StationsPD()
        self.stations_t = StationsT()
        self.stations_et = StationsET()
        self.stations_p = StationsP()
        self.stations = [
            self.stations_pd,
            self.stations_t,
            self.stations_et,
            self.stations_p]

        self._is_active = False

        self._db_schema_valid = None

    def _check_paras(self, paras, valid_paras=["p_d", "p", "t", "et"]):
        valid_paras = ["p_d", "p", "t", "et"]
        for para in paras:
            if para not in valid_paras:
                raise ValueError(
                    "The given parameter {para} is not valid.".format(
                        para=para))

    @property
    def _alembic_config(self):
        from .alembic.config import config as alembic_config
        return alembic_config

    @db_engine.deco_is_superuser
    def create_db_schema(self, if_exists=None, silent=False, owner=None):
        """Create the database schema.

        Parameters
        ----------
        if_exists : str, optional
            What to do if the tables already exist.
            If None the user gets asked.
            If "D" or "drop" the tables get dropped and recreated.
            If "I" or "ignore" the existing tables get ignored and the creation of the schema continues for the other.
            If "E" er "exit" the creation of the schema gets exited.
            The default is None.
        silent : bool, optional
            If True the user gets not asked if the tables already exist.
            If True, if_exists must not be None.
            The default is False.
        owner : str, optional
            The user that should get the ownership of the tables and schemas.
            If None the current database user will be the owner.
            The default is None.
        """
        # check silent
        if silent and if_exists is None:
            raise ValueError("silent can only be True if if_exists is not None.")

        # add POSTGIS extension
        with db_engine.connect() as con:
            con.execute(sqltxt("CREATE EXTENSION IF NOT EXISTS postgis;"))
            con.commit()

        # check for existing tables
        from .db.models import ModelBase
        with db_engine.connect() as con:
            res = con.execute(sqltxt("""
                SELECT table_name
                FROM information_schema.tables
                WHERE table_schema = 'public' AND table_type = 'BASE TABLE';""")
                ).fetchall()
            existing_tables = [table[0] for table in res]
            problem_tables = [table.name
                              for table in ModelBase.metadata.tables.values()
                              if table.name in existing_tables]
        if len(problem_tables)>0 and silent:
            log.info("The following tables already exist on the database:\n - " +
                  "\n - ".join([table for table in problem_tables]))

            # ask the user what to do if if_exists is None
            if if_exists is None:
                print(textwrap.dedent(
                    """What do you want to do?
                    - [D] : Drop all the tables and recreate them again.
                    - [I] : Ignore those tables and continue with the creation of the schema.
                    - [E] : Exit the creation of the schema."""))
                while True:
                    answer = input("Your choice: ").upper()
                    if answer in ["D", "E", "I"]:
                        if_exists = answer
                        break
                    else:
                        print("Please enter a valid answer.")

            # execute the choice
            if if_exists.upper() == "D":
                log.info("Dropping the tables.")
                with db_engine.connect() as con:
                    for table in problem_tables:
                        con.execute(sqltxt(f"DROP TABLE {table} CASCADE;"))
                    con.commit()
            elif if_exists.upper() == "E":
                log.info("Exiting the creation of the schema.")
                return

        # create the tables
        log.info("Creating the tables and views.")
        with db_engine.connect() as con:
            ModelBase.metadata.create_all(con, checkfirst=True)
            con.commit()

        # create the schema for the timeseries
        with db_engine.connect() as con:
            con.execute(sqltxt("CREATE SCHEMA IF NOT EXISTS timeseries;"))
            con.commit()

        # set the owner of the tables
        if owner is not None:
            log.info(f"Setting the owner of the tables to {owner}.")
            with db_engine.connect() as con:
                for table in ModelBase.metadata.tables.values():
                    con.execute(sqltxt(f"ALTER TABLE {table.name} OWNER TO {owner};"))
                con.execute(sqltxt(f"ALTER SCHEMA timeseries OWNER TO {owner};"))
                con.commit()

        # tell alembic that the actual database schema is up-to-date
        from alembic import command
        command.stamp(self._alembic_config, "head")

    def upgrade_db_schema(self, revision="head"):
        """Upgrade the database schema to a specific revision.

        Parameters
        ----------
        revision : str, optional
            The revision to upgrade to.
            If "head" the database gets upgraded to the latest revision.
            The default is "head".
        """
        from alembic import command
        from alembic.runtime.environment import EnvironmentContext
        from alembic.script import ScriptDirectory

        from .db.models import ModelBase

        # get alembic context
        context = EnvironmentContext(
            self._alembic_config,
            ScriptDirectory(self._alembic_config.get_main_option("script_location")))
        with db_engine.connect() as con:
            context.configure(connection=con)
            migration_context = context.get_context()
            pre_version = pv.parse(migration_context.get_current_revision())
            head_version = pv.parse(context.get_head_revision())

            # check revision
            if revision == "head":
                version = head_version
            else:
                version = pv.parse(revision)

            # remove all views
            for view in ModelBase.metadata.views:
                view.drop_view(None, con)

            # apply the migrations
            if version > pre_version:
                command.upgrade(self._alembic_config, revision)
            elif version < pre_version:
                command.downgrade(self._alembic_config, revision)

            # check if revision is the same as head
            if "head" in revision or version == head_version:
                # create the views
                for view in ModelBase.metadata.views:
                    view.create_view(None, con)
            else:
                log.info("The views are not created because the revision is not head. This can resolve in errors during module execution.")

    def _check_db_schema(self):
        """Check the database schema for differences to the models.

        Returns
        -------
        bool
            Whether the database schema is up-to-date.

        Raises
        ------
        Exception
            If the database schema is not up-to-date.
        """
        if self._db_schema_valid:
            return self._db_schema_valid
        else:
            from alembic import command
            try:
                command.check(self._alembic_config)
                self._db_schema_valid = True
            except Exception as e:
                self._db_schema_valid = False
                if "Target database is not up to date" in str(e):
                    note = "You may need to run `weatherDB upgrade-db-schema` in CLI or use `weatherDB.Broker().upgrade_db_schema()` as python script to upgrade the database schema first."
                    if sys.version_info >= (3, 11):
                        e.add_note(note)
                    else:
                        raise Exception(str(e) + "\n" + note) from e

                raise e

    @db_engine.deco_all_privileges
    def initiate_db(self, **kwargs):
        """Initiate the Database.

        Downloads all the data from the CDC server for the first time.
        Updates the multi-annual data and the richter-class for all the stations.
        Quality checks and fills up the timeseries.

        Parameters
        ----------
        **kwargs : dict
            The keyword arguments to pass to the called methods of the stations
        """
        log.info("Broker initiate_db starts")
        self._check_db_schema()

        with self.activate():
            self.update_meta(
                paras=["p_d", "p", "t", "et"], **kwargs)
            self.update_raw(
                paras=["p_d", "p", "t", "et"],
                only_new=False,
                **kwargs)
            self.update_ma_raster(
                paras=["p_d", "p", "t", "et"],
                **kwargs)
            self.stations_p.update_richter_class(**kwargs)
            self.quality_check(paras=["p", "t", "et"], **kwargs)
            self.fillup(paras=["p", "t", "et"], **kwargs)
            self.richter_correct(**kwargs)

            self.set_db_version()

            self.vacuum()

    def update_raw(self, only_new=True, paras=["p_d", "p", "t", "et"], **kwargs):
        """Update the raw data from the DWD-CDC server to the database.

        Parameters
        ----------
        only_new : bool, optional
            Get only the files that are not yet in the database?
            If False all the available files are loaded again.
            The default is True.
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to the update_raw method of the stations
        """
        self._check_paras(paras)
        log.info("Broker update_raw starts")
        self._check_db_schema()

        with self.activate():
            for stations in self.stations:
                if stations._para in paras:
                    stations.update_raw(only_new=only_new, **kwargs)

    def update_meta(self, paras=["p_d", "p", "t", "et"], **kwargs):
        """Update the meta file from the CDC Server to the Database.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to update_meta method of the stations
        """
        self._check_paras(paras)
        log.info("Broker update_meta starts")
        self._check_db_schema()

        with self.activate():
            for stations in self.stations:
                if stations._para in paras:
                    stations.update_meta(**kwargs)

    def update_ma_raster(self, paras=["p_d", "p", "t", "et"], **kwargs):
        """Update the multi-annual data from raster to table.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to update_ma_raster method of the stations
        """
        self._check_paras(paras)
        log.info("Broker update_ma_raster starts")
        self._check_db_schema()

        with self.activate():
            for stations in self.stations:
                if stations._para in paras:
                    stations.update_ma_raster(**kwargs)

    def update_period_meta(self, paras=["p_d", "p", "t", "et"], **kwargs):
        """Update the periods in the meta table.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to update_period_meta method of the stations
        """
        self._check_paras(paras=paras,
                          valid_paras=["p_d", "p", "t", "et"])
        log.info("Broker update_period_meta starts")
        self._check_db_schema()

        with self.activate():
            for stations in self.stations:
                if stations._para in paras:
                    stations.update_period_meta(**kwargs)

    def quality_check(self, paras=["p", "t", "et"], with_fillup_nd=True, **kwargs):
        """Do the quality check on the stations raw data.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p", "t", "et"].
            The default is ["p", "t", "et"].
        with_fillup_nd : bool, optional
            Should the daily precipitation data get filled up if the 10 minute precipitation data gets quality checked.
            The default is True.
        **kwargs : dict
            The keyword arguments to pass to quality_check method of the stations
        """
        self._check_paras(
            paras=paras,
            valid_paras=["p", "t", "et"])
        log.info("Broker quality_check starts")
        self._check_db_schema()

        with self.activate():
            if with_fillup_nd and "p" in paras:
                self.stations_pd.fillup(**kwargs)

            for stations in self.stations:
                if stations._para in paras:
                    stations.quality_check(**kwargs)

    def last_imp_quality_check(self, paras=["p", "t", "et"], with_fillup_nd=True, **kwargs):
        """Quality check the last imported data.

        Also fills up the daily precipitation data if the 10 minute precipitation data should get quality checked.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p", "t", "et"].
            The default is ["p", "t", "et"].
        with_fillup_nd : bool, optional
            Should the daily precipitation data get filled up if the 10 minute precipitation data gets quality checked.
            The default is True.
        **kwargs : dict
            The keyword arguments to pass to last_imp_quality_check method of the stations.
            If with_fillup_nd is True, the keyword arguments are also passed to the last_imp_fillup method of the stations_pd.
        """
        self._check_paras(
                paras=paras,
                valid_paras=["p", "t", "et"])
        log.info("Broker last_imp_quality_check starts")
        self._check_db_schema()

        with self.activate():
            if with_fillup_nd and "p" in paras:
                self.stations_pd.last_imp_fillup(**kwargs)

            for stations in self.stations:
                if stations._para in paras:
                    stations.last_imp_quality_check(**kwargs)

    def fillup(self, paras=["p", "t", "et"], **kwargs):
        """Fillup the timeseries.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to fillup method of the stations
        """
        log.info("Broker fillup starts")
        self._check_db_schema()

        with self.activate():
            self._check_paras(paras)
            for stations in self.stations:
                if stations._para in paras:
                    stations.fillup(**kwargs)

    def last_imp_fillup(self, paras=["p", "t", "et"], **kwargs):
        """Fillup the last imported data.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to last_imp_fillup method of the stations
        """
        log.info("Broker last_imp_fillup starts")
        self._check_db_schema()

        with self.activate():
            self._check_paras(paras)
            for stations in self.stations:
                if stations._para in paras:
                    stations.last_imp_fillup(**kwargs)

    def richter_correct(self, **kwargs):
        """Richter correct all of the precipitation data.

        Parameters
        ----------
        **kwargs : dict
            The keyword arguments to pass to richter_correct method of the stations_p
        """
        log.info("Broker: last_imp_corr starts")
        self._check_db_schema()

        with self.activate():
            self.stations_p.richter_correct(**kwargs)

    def last_imp_corr(self, **kwargs):
        """Richter correct the last imported precipitation data.

        Parameters
        ----------
        **kwargs : dict
            The keyword arguments to pass to last_imp_corr method of the stations
        """
        log.info("Broker: last_imp_corr starts")
        self._check_db_schema()

        with self.activate():
            self.stations_p.last_imp_corr(**kwargs)

    def update_db(self, paras=["p_d", "p", "t", "et"], **kwargs):
        """The regular Update of the database.

        Downloads new data.
        Quality checks the newly imported data.
        Fills up the newly imported data.

        Parameters
        ----------
        paras : list of str, optional
            The parameters for which to do the actions.
            Can be one, some or all of ["p_d", "p", "t", "et"].
            The default is ["p_d", "p", "t", "et"].
        **kwargs : dict
            The keyword arguments to pass to the called methods of the stations
        """
        log.info("Broker update_db starts")
        self._check_paras(paras)
        self._check_db_schema()

        with self.activate():
            if self.get_db_version() is None or pv.parse(__version__) > self.get_db_version():
                log.info("--> There is a new version of the python script. Therefor the database is recalculated completly")
                self.initiate_db()
            else:
                self.update_meta(paras=paras, **kwargs)
                self.update_raw(paras=paras, **kwargs)
                if "p_d" in paras:
                    paras.remove("p_d")
                self.last_imp_quality_check(paras=paras, **kwargs)
                self.last_imp_fillup(paras=paras, **kwargs)
                self.last_imp_corr(**kwargs)

    def vacuum(self, do_analyze=True):
        sql = "VACUUM {analyze};".format(
            analyze="ANALYZE" if do_analyze else "")
        with db_engine.connect() as con:
            con.execute(sqltxt(sql))

    def get_setting(self, key):
        """Get a specific settings value from the database.

        Parameters
        ----------
        key : str
            The key of the setting.

        Returns
        -------
        value: str or None
            The database settings value.
        """
        with db_engine.connect() as con:
            res = con.execute(
                sqltxt(f"SELECT value FROM settings WHERE key='{key}';")
                ).fetchone()
        if res is None:
            return None
        else:
            return res[0]

    def set_setting(self, key:str, value:str):
        """Set a specific setting.

        Parameters
        ----------
        key : str
            The key of the setting.
        value : str
            The value of the setting.
        """
        with db_engine.connect() as con:
            con.execute(sqltxt(
                f"""INSERT INTO settings
                VALUES ('{key}', '{value}')
                ON CONFLICT (key)
                    DO UPDATE SET value=EXCLUDED.value;"""))
            con.commit()

    def get_db_version(self):
        """Get the package version that the databases state is at.

        Returns
        -------
        version
            The version of the database.
        """
        res = self.get_setting("version")
        if res is not None:
            res = pv.parse(res)
        return res

    def set_db_version(self, version=pv.parse(__version__)):
        """Set the package version that the databases state is at.

        Parameters
        ----------
        version: pv.Version, optional
            The Version of the python package
            The default is the version of this package.
        """
        if not isinstance(version, pv.Version):
            raise TypeError("version must be of type pv.Version")
        self.set_setting("version", str(version))

    @property
    def is_any_active(self):
        """Get the state of the broker.

        Returns
        -------
        bool
            Whether the broker is active.
        """
        return self.get_setting("broker_active") == "True"

    @property
    def is_active(self):
        """Get the state of the broker.

        Returns
        -------
        bool
            Whether the broker is active.
        """
        return self._is_active

    @is_active.setter
    def is_active(self, is_active:bool):
        """Set the state of the broker.

        Parameters
        ----------
        is_active : bool
            Whether the broker is active.
        """
        self._is_active = is_active
        self.set_setting("broker_active", str(is_active))

    def _deactivate(self):
        self.is_active = False

    @contextmanager
    def activate(self):
        """Activate the broker in a context manager."""
        try:
            if self.is_any_active and not self.is_active:
                raise RuntimeError("Another Broker is active and therefor this broker is not allowed to run.")
            self.is_active = True
            atexit.register(self._deactivate)
            yield self
        finally:
            self._deactivate()
            atexit.unregister(self._deactivate)


from alembic import context
import re
from setuptools_scm import get_version
from packaging.version import parse as vparse

import weatherDB as wdb
from weatherDB.db.models import ModelBase
from weatherDB.db.connections import db_engine

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.
config = context.config

# add your model's MetaData object here
# for 'autogenerate' support
# in console do: weatherDB>alembic -c alembic\alembic.ini revision --autogenerate -m "comment" --rev-id "V1.0.0"
target_metadata = ModelBase.metadata

# check for alembic database copnnection in the weatherDB config
# ##############################################################
engine = config.attributes.get("engine", None)
if engine is None and \
        wdb.config.has_option('alembic', 'connection') and \
        wdb.config.has_section(f'database:{wdb.config.get("alembic", "connection")}'):
    print("Setting the database connection to the users configuration of 'alembic.connection'.")
    wdb.config.set(
        'database',
        'connection',
        wdb.config.get("alembic", "connection"))
    engine = db_engine.engine

# get other values from config
# ############################
exclude_tables = re.sub(
    r"\s+",
    '',
    config.get_main_option('exclude_tables', '')
).split(',')
valid_schemas = ModelBase.metadata._schemas
valid_tables = {
    schema: [table.name
             for table in ModelBase.metadata.tables.values()
             if table.schema == schema]
    for schema in ModelBase.metadata._schemas}

def include_name(name, type_, parent_names, *args,**kwargs):
    if type_ == "schema":
        return (name in valid_schemas) or (name is None)
    else:
        schema = parent_names["schema_name"] if parent_names["schema_name"] is not None else "public"
        if schema not in valid_schemas:
            return False

        if type_ == "table":
            return name in valid_tables.get(schema, []) and name not in exclude_tables

        table = parent_names["table_name"]
        return table in valid_tables.get(schema, []) and table not in exclude_tables


# get revision id
# ###############
def process_revision_directives(context, revision, directives):
    # extract Migration
    migration_script = directives[0]

    # get version from setuptools_scm or weatherDB if not in git
    try:
        version = vparse(get_version(root="..", relative_to=wdb.__file__))
    except LookupError:
        version = vparse(wdb.__version__)
    migration_script.rev_id = f"V{version.base_version}"


# migration functions
# ###################
def run_migrations_online() -> None:
    """Run migrations in 'online' mode.

    In this scenario we need to create an Engine
    and associate a connection with the context.

    """
    with engine.connect() as connection:
        context.configure(
            connection=connection,
            target_metadata=target_metadata,
            include_schemas=True,
            include_name=include_name,
            process_revision_directives=process_revision_directives,
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    raise NotImplementedError("offline mode is not supported")
else:
    run_migrations_online()



import click
import sys
from pathlib import Path
from textwrap import dedent


sys.path.insert(0, Path(__file__).resolve().parent.parent.as_posix())
import weatherDB

# main cli group
# ---------------------------------------

@click.group(help="This is the Command line interface of the weatherDB package.",
             chain=True,
             context_settings=dict(
                 show_default=True
                 )
            )
@click.option('--do-logging/--no-logging',
              is_flag=True, default=True, show_default=True,
              help="Should the logging be done to the console?")
@click.option('--connection', '-c',
              type=str, default=None, show_default=False,
              help="The connection to use. Default is the value from the configuration file.")
@click.option('--verbose', '-v', # defined here but used in safe_cli
              is_flag=True, default=False, show_default=True,
              help="Should the verbose mode be activated? -> This will print complete tracebacks on errors.")
def cli(do_logging, connection, verbose):
    if do_logging:
        click.echo("logging to console is set on")
        handlers = weatherDB.config.getlist("logging", "handler")
        if "console" not in handlers:
            handlers.append("console")
            weatherDB.config.set("logging", "handler", handlers)

    if connection is not None:
        print(f"setting the connection to {connection}")
        weatherDB.config.set("database", "connection", connection)


def safe_entry():
    try:
        cli()
    except Exception as e:
        if ("-v" in sys.argv) or ("--verbose" in sys.argv):
            raise e
        notes = "\n" + '\n'.join(e.__notes__) if hasattr(e, '__notes__') else ''
        click.echo(f"\033[31;1;4mAn error occurred: {e}{notes}\033[0m", err=True)
        sys.exit(1)

# cli statements to initialize the module
# ---------------------------------------

@cli.command(short_help="Create the database schema for the first time.")
@click.option('--owner', '-o',
              type=str, default=None,
              help=dedent("""
                 The user that should get the ownership of the created tables and schemas.
                 As default is the current user will be the owner."""))
def create_db_schema(owner):
    click.echo("starting to create database schema")
    broker = weatherDB.broker.Broker()
    broker.create_db_schema(owner=owner)


@cli.command(short_help="Use Alembic directly to update or downgrade the database schema. Have a look at the official alembic documentation for more information.")
@click.option('--revision', '-r',
              type=str, default="head", show_default=True,
              help="The revision ID (weatherDB Version e.g. 'V1.0.2') to upgrade/downgrade to.")
def upgrade_db_schema(revision):
    click.echo("starting to upgrade database schema")
    broker = weatherDB.broker.Broker()
    broker.upgrade_db_schema(revision=revision)


@cli.command(short_help="Create User configuration file.")
@click.option('--file', '-f',
              type=click.Path(), default="ask", show_default=True,
              help="The file to save the user configuration to.")
@click.option('--on-exists', '-e',
              type=str, default="ask", show_default=True,
              help="What to do if the file already exists. Options are 'ask', 'overwrite', 'define' or 'error'.")
def create_user_config(file, on_exists):
    weatherDB.config.create_user_config(user_config_file=file, on_exists=on_exists)


@cli.command(short_help="Download the needed multi-annual raster data from zenodo to the data folder.")
@click.option('--overwrite', '-o',
              type=bool, default=None, show_default=True,
              help="Should the multi annual rasters be downloaded even if they already exist?")
@click.option('--which', '-w',
              type=str, default=["all"], show_default=True,
              multiple=True,
              help="Which raster to download. Options are 'dwd', 'hyras', 'regnie' or 'all'.")
@click.option("--update-user-config", "-u",
              type=bool, default=False, show_default=True, is_flag=True,
              help="Should the user configuration be updated with the path to the downloaded rasters?")
def download_ma_rasters(which, overwrite, update_user_config):
    """Get the multi annual rasters on which bases the regionalisation is done.

    The refined multi annual datasets, that are downloaded are published on Zenodo:
    Schmit, M.; Weiler, M. (2023). German weather services (DWD) multi annual meteorological rasters for the climate period 1991-2020 refined to 25m grid (1.0.0) [Data set]. Zenodo. https://doi.org/10.5281/zenodo.10066045
    """
    click.echo("starting downloading multi annual raster data")
    from weatherDB.utils.get_data import download_ma_rasters
    download_ma_rasters(overwrite=overwrite)


@cli.command(short_help="Download the needed digital elevation model raster data from Copernicus to the data folder.")
@click.option('--overwrite/--no-overwrite', '-o/-no-o',
              type=bool, is_flag=True, default=None, show_default=False,
              help="Should the digital elevation model raster be downloaded even if it already exists?")
@click.option('--extent', '-e',
              type=tuple, default=(5.3, 46.1, 15.6, 55.4), show_default=True,
              help="The extent in WGS84 of the DEM data to download. The default is the boundary of germany + ~40km.")
@click.option("--update-user-config", "-u",
              type=bool, default=False, show_default=True, is_flag=True,
              help="Should the user configuration be updated with the path to the downloaded DEM?")
def download_dem(overwrite, extent):
    """Download the newest DEM data from the Copernicus Sentinel dataset.

    Only the GLO-30 DEM, wich has a 30m resolution, is downloaded as it is freely available.
    If you register as a scientific researcher also the EEA-10, with 10 m resolution, is available.
    You will have to download the data yourself and define it in the configuration file.

    After downloading the data, the files are merged and saved as a single tif file in the data directory in a subfolder called 'dems'.
    To use the DEM data in the WeatherDB, you will have to define the path to the tif file in the configuration file.

    Source:
    Copernicus DEM - Global and European Digital Elevation Model. Digital Surface Model (DSM) provided in 3 different resolutions (90m, 30m, 10m) with varying geographical extent (EEA: European and GLO: global) and varying format (INSPIRE, DGED, DTED). DOI:10.5270/ESA-c5d3d65.
    """
    click.echo("Starting downloading digital elevation model from Copernicus")
    from weatherDB.utils.get_data import download_dem
    download_dem(overwrite=overwrite, extent=extent)


# cli statements to update the database
# ---------------------------------------
@cli.command(short_help="Update the complete database. Get the newest data from DWD and treat it.")
def update_db():
    click.echo("starting updating the database")
    broker = weatherDB.broker.Broker()
    broker.update_db()


@cli.command(short_help="Update the meta data in the database. Get the newest meta data from DWD.")
def update_meta():
    click.echo("updating the meta data")
    broker = weatherDB.broker.Broker()
    broker.update_meta()


@cli.command(short_help="Update the Richter classes of the precipitation stations in the database.")
def update_richter_class():
    click.echo("starting updating the regionalisation")
    weatherDB.StationsP().update_richter_class()


@cli.command(short_help="Update the multi annual raster values in the database.")
def update_ma_raster():
    click.echo("starting updating the multi annual raster data")
    broker = weatherDB.broker.Broker()
    broker.update_ma_raster()


@cli.command(short_help="Update the raw data of the complete database.")
def update_raw():
    click.echo("starting updating the raw data")
    broker = weatherDB.broker.Broker()
    broker.update_raw()


@cli.command(short_help="Do the quality check of the complete database.")
def quality_check():
    click.echo("starting quality check")
    broker = weatherDB.broker.Broker()
    broker.quality_check()


@cli.command(short_help="Do the filling of the complete database.")
def fillup():
    click.echo("starting filling up")
    broker = weatherDB.broker.Broker()
    broker.fillup()


@cli.command(short_help="Do the richter correction of the complete database.")
def richter_correct():
    click.echo("starting richter correction")
    broker = weatherDB.broker.Broker()
    broker.richter_correct()


# cli admin stuff
# ---------------------------------------

@cli.command(short_help="Set the db version to the current weatherDB version to prevent recalculation of the whole database. (!!!Only use this if you're sure that the database did all the necessary updates!!!)")
def set_db_version():
    click.echo(dedent(
        """Are you sure you want to set the db version to the current weatherDB version?
           This will prevent the recalculation of the whole database if there was an update and could resolve in old values in the database."""))
    if click.confirm("Are you sure you want to continue?"):
        click.echo("starting setting db version")
        broker = weatherDB.broker.Broker()
        broker.set_db_version()
    else:
        click.echo("aborting setting db version")


# cli
# ---------------------------------------
if __name__=="__main__":
    safe_entry()




import re
from pandas import Timestamp, NaT, Timedelta
import datetime

class TimestampPeriod(object):
    """A class to save a Timespan with a minimal and maximal Timestamp.
    """
    _COMPARE = {
        "inner": {
            0: max,
            1: min},
        "outer": {
            0: min,
            1: max}}
    _REGEX_HAS_TIME = re.compile(
        r"((^\d{6}[ \-\.]+)"+ # 991231
        r"|(^\d{8}[ \-\.]*)|" + # 19991231
        r"(^\d{1,4}[ \-\.]\d{1,2}[ \-\.]\d{1,4}[ \-\.]+))" + # 1999-12-31
        r"+(\d+)") # has additional numbers -> time

    def __init__(self, start, end, tzinfo="UTC"):
        """Initiate a TimestampPeriod.

        Parameters
        ----------
        start : pd.Timestamp or similar
            The start of the Period.
        end : pd.Timestamp or similar
            The end of the Period.
        tzinfo : str or datetime.timezone object or None, optional
            The timezone to set to the timestamps.
            If the timestamps already have a timezone they will get converted.
            If None, then the timezone is not changed or set.
            The default is "UTC".
        """
        # check if input is a date or a timestamp
        if ((isinstance(start, datetime.date) and isinstance(end, datetime.date)) or
            (isinstance(start, str) and not self._REGEX_HAS_TIME.match(start) and
             isinstance(end, str) and not self._REGEX_HAS_TIME.match(end))):
            self.is_date = True
        else:
            self.is_date = False

        # convert to correct timestamp format
        period = list([start, end])
        for i, tstp in enumerate(period):
            if not isinstance(tstp, Timestamp):
                period[i] = Timestamp(tstp)

            # check timezone
            self.tzinfo = tzinfo
            if tzinfo is not None:
                if period[i].tzinfo is None:
                    period[i] = period[i].tz_localize(tzinfo)
                else:
                    period[i] = period[i].tz_convert(tzinfo)

        self.start = period[0]
        self.end = period[1]

    @staticmethod
    def _check_period(period):
        if isinstance(period, TimestampPeriod):
            period = TimestampPeriod(*period)
        return period

    def union(self, other, how="inner"):
        """Unite 2 TimestampPeriods to one.

        Compares the Periods and computes a new one.

        Parameters
        ----------
        other : TimestampPeriod
            The other TimestampPeriod with whome to compare.
        how : str, optional
            How to compare the 2 TimestampPeriods.
            Can be "inner" or "outer".
            "inner": the maximal Timespan for both is computed.
            "outer": The minimal Timespan for both is computed.
            The default is "inner".

        Returns
        -------
        TimestampPeriod
            A new TimespanPeriod object uniting both TimestampPeriods.
        """
        other = self._check_period(other)

        # check for daily period in elements
        tdsother = [Timedelta(0), Timedelta(0)]
        tdsself = [Timedelta(0), Timedelta(0)]
        if self.is_date and not other.is_date and not other.is_empty():
            tdsself[1] = Timedelta(
                hours=other.end.hour,
                minutes=other.end.minute,
                seconds=other.end.second)
        elif not self.is_date and other.is_date and not self.is_empty():
            tdsother[1] = Timedelta(
                hours=self.end.hour,
                minutes=self.end.minute,
                seconds=self.end.second)

        # check if empty and inner
        if how=="inner" and (self.is_empty() or other.is_empty()):
            return TimestampPeriod(None, None)

        # get the united period
        period = [None, None]
        for i in range(2):
            comp_list = [val + td
                for val, td in zip([self[i], other[i]],
                                   [tdsself[i], tdsother[i]])
                if isinstance(val, Timestamp)]
            if len(comp_list) > 0:
                period[i] = self._COMPARE[how][i](comp_list)

        # check if end < start
        if period[0]>=period[1]:
            period = (None, None)

        # return the period
        if self.is_date and other.is_date and all(period):
            # if both were data periods, then the result should also be a date period
            return TimestampPeriod(
                *[val.date() for val in period
                             if isinstance(val, Timestamp)])
        else:
            return TimestampPeriod(*period)

    def get_period(self):
        return (self.start, self.end)

    def __getitem__(self, key):
        if key == 0 or key == "start":
            return self.start
        elif key == 1 or key == "end":
            return self.end

    def __setitem__(self, key, value):
        if value != Timestamp:
            value = Timestamp(value)

        if key == 0 or key == "start":
            self.start = value
        elif key == 1 or key == "end":
            self.end = value

    def __iter__(self):
        return self.get_period().__iter__()

    def __str__(self):
        msg = "TimestampPeriod: {0} - {1}"
        if self.is_date:
            return msg.format(*self.strftime(format="%Y-%m-%d"))
        else:
            return msg.format(*self.strftime(format="%Y-%m-%d %H:%M:%S"))

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        other = self._check_period(other)

        if self.start == other.start and self.end == other.end:
            return True
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __lt__(self, other):
        return self.inside(other)

    def __le__(self, other):
        return self.inside(other)

    def __gt__(self, other):
        return self.contains(other)

    def __ge__(self, other):
        return self.contains(other)

    def has_NaT(self):
        """Has the TimestampPeriod at least one NaT.

        This means that the start or end is not given.
        Normally this should never happen, because it makes no sense.

        Returns
        -------
        bool
            True if the TimestampPeriod has at least on NaT.
            False if the TimestampPeriod has at least a start or a end.
        """
        return any([tstp is NaT for tstp in self])

    def has_only_NaT(self):
        """Has the TimestampPeriod only NaT, meaning is empty.

        This means that the start and end is not given.

        Returns
        -------
        bool
            True if the TimestampPeriod is empty.
            False if the TimestampPeriod has a start and an end.
        """
        return all([tstp is NaT for tstp in self])

    def is_empty(self):
        """Is the TimestampPeriod empty.

        This means that the start and end is not given.

        Returns
        -------
        bool
            True if the TimestampPeriod is empty.
            False if the TimestampPeriod has a start and an end.
        """
        return self.has_only_NaT()

    def strftime(self, format="%Y-%m-%d %H:%M:%S"):
        """Convert the TimestampPeriod to a list of strings.

        Formates the Timestamp as a string.

        Parameters
        ----------
        format : str, optional
            The Timestamp-format to use.
            The Default is "%Y-%m-%d %H:%M:%S"

        Returns
        -------
        list of 2 strings
            A list of the start and end of the TimestampPeriod as formated string.
        """
        out = [tstp.strftime(format) if tstp is not NaT else None
                    for tstp in self.get_period()]
        return out

    def inside(self, other):
        """Is the TimestampPeriod inside another TimestampPeriod?

        Parameters
        ----------
        other : Timestampperiod or tuple of 2 Timestamp or Timestamp strings
            The other Timestamp to test against.
            Test if this TimestampPeriod is inside the other.

        Returns
        -------
        bool
            True if this TimestampPeriod is inside the other.
            Meaning that the start is higher or equal than the others starts
            and the end is smaller than the others end.
        """
        other = self._check_period(other)
        if self.start >= other.start and self.end <= other.end:
            return True
        else:
            return False

    def contains(self, other):
        """Does this TimestampPeriod contain another TimestampPeriod?

        Parameters
        ----------
        other : Timestampperiod or tuple of 2 Timestamp or Timestamp strings
            The other Timestamp to test against.
            Test if this TimestampPeriod contains the other.

        Returns
        -------
        bool
            True if this TimestampPeriod contains the other.
            Meaning that the start is smaller or equal than the others starts
            and the end is higher than the others end.
        """
        other = self._check_period(other)
        return other.inside(self)

    def get_sql_format_dict(self, format="'%Y%m%d %H:%M'"):
        """Get the dictionary to use in sql queries.

        Parameters
        ----------
        format : str, optional
            The Timestamp-format to use.
            The Default is "'%Y%m%d %H:%M'"

        Returns
        -------
        dict
            a dictionary with 2 keys (min_tstp, max_tstp) and the corresponding Timestamp as formated string.
        """
        period_str = self.strftime(format=format)
        period_str = [str(el).replace("None", "NULL") for el in period_str]
        return dict(min_tstp=period_str[0], max_tstp=period_str[1])

    def get_interval(self):
        """Get the interval of the TimestampPeriod.

        Returns
        -------
        pd.Timedelta
            The interval of this TimestampPeriod.
            E.G. Timedelta(2 days 12:30:12)
        """
        return self.end - self.start

    def get_middle(self):
        """Get the middle Timestamp of the TimestampPeriod.

        Returns
        -------
        Timestamp
            The middle Timestamp of this TimestampPeriod.
        """
        middle = self.start + self.get_interval() / 2
        if self.is_date:
            middle = Timestamp(middle.date())
        if self.tzinfo is not None:
            if middle.tzinfo is None:
                middle = middle.tz_localize(self.tzinfo)
        return middle

    def copy(self):
        """Copy this TimestampPeriod.

        Returns
        -------
        TimestampPeriod
            a new TimestampPeriod object that is equal to this one.
        """
        new = TimestampPeriod(self.start, self.end)
        new.is_date = self.is_date
        return new

    def expand_to_timestamp(self):
        if self.is_date:
            return TimestampPeriod(
                start=self.start,
                end=self.end + Timedelta(
                    hours=23, minutes=59, seconds=59, milliseconds=999))
        else:
            return self


